/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/protocol_crypt.h"
#include <stdio.h>
#include <stdlib.h>
#include <boost/format.hpp>

// ##################################################################
/**
* @brief Initialisieren der Membervariablen.
*
* @param connectionId Nummer der Kanalklasse fuer z.B. Trace.
* @param deviceAddress Busnummer der Kanalklasse.
*/
Datafox::Communication::Protocol_crypt::Protocol_crypt( int connection_id, uint8_t device_address )
    : Protocol_base( connection_id, device_address ) 
{
    // Umgebung zuruecksetzen
    reset();

    // Zufallsgenerator starten
    std::srand( (unsigned int)std::time( nullptr ) );

    // Halten der vorhergehenden Sendedaten, um bei Wiederholdungen diese wiederzuverwenden.
    m_packet_number_previous = 0xFF;
    memset( m_packet_data_previous, 0, sizeof(m_packet_data_previous) );
    m_packet_data_size_previous = 0;
}

// ##################################################################
/**
* @brief Destruktor.
*
*/
Datafox::Communication::Protocol_crypt::~Protocol_crypt() 
{
}

//#########################################################
//####################### HELPER ##########################
//#########################################################

// ##################################################################
/**
* @brief Initialisiert die grundlegend benoetigten Membervariablen.
*/
void Datafox::Communication::Protocol_crypt::reset()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    Protocol_base::reset();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceAddress Anzusprechende Geraetenummer.
*/
// Modulbefehl senden und Antwort zurueckgeben
//  deviceAddress -> Adresse (Geraetenummer) des Empfaengers
//  module   -> Nummer des anzusprechenden Modules
//  func -> Nummer der Funktion im Module
//	buf	     -> Buffer fuer Funktionsparameter.
//	count    -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
//
//  return -> <0 -> Kommunikationsfehler
//             0 -> senden und empfangen und ausfuehren erfolgreich
//			  >0 -> module-/Funktionsfehler
int Datafox::Communication::Protocol_crypt::call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int context_value )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u, %p, %u, %i )\n", __FILE__, __LINE__, __func__, context, module, func, buffer, count, context_value );
    int repeat = 0;

    // Im Crypt-Mode Userpassword muss auch ein solches vorliegen!
    // Bei Ping-Befehl pruefung unterlassen.
    if ( (context == IV_CONTEXT_MASTER)
        && !((module == IV_MODULE_PROTOCOL) && (func == IV_FUNCTION_PING))
        && aes_crypt_session().m_crypt_mode == IV_PACKET_CRYPT_MODE_PASSPHRASE && !aes_crypt_session().is_valid_user_key() )
    {
        return -IV_ERROR_NEED_USER_PASSWORD;
    }

label_again:
    // Muss eine Schluesselaushandlung aufgrund folgender Faelle durchgefuehrt werden:
    // Es ist eine Anfrage (ausser Schluesselaustausch selbst) die verschluesselte Kommunikation wird unterstuetzt und
    // soll eingesetzt werden und 
    // 1. Es wurde noch kein Session-Key ausgehandelt.
    // 2. Der aktuelle Session-Key ist aufgrund eines Timeouts ungueltig.
    if ( context == IV_CONTEXT_MASTER
        && !(module == IV_MODULE_PROTOCOL_CRYPT && func == IV_FUNCTION_EXCHANGE_PUBLIC_KEY) && aes_crypt_session().m_crypt_mode > 1 
        && (aes_crypt_session().m_session_time == 0 || (aes_crypt_session().m_session_timeout && time( nullptr ) > (aes_crypt_session().m_session_time + aes_crypt_session().m_session_timeout))) )
    {
        if ( aes_crypt_session().m_session_time )
        {
            trace( "%s(%i):WARNING(1):  %s() [ session t(time): %i, t(timeout): %i, t(time-timeout): %i ]\n", __FILE__, __LINE__, __func__, static_cast<int>( time( nullptr ) - aes_crypt_session().m_session_time ), aes_crypt_session().m_session_timeout, static_cast<int>( time( NULL ) - (aes_crypt_session().m_session_time + aes_crypt_session().m_session_timeout) ) );
        }

        // Aushandlung bereits durchgefuehrt und laenger als 9 Sekunden abgelaufen?
        // Dann muessen die Crypt-Routinen neu aufgesetzt werden!
        if ( aes_crypt_session().m_session_time > 0 && (time( nullptr ) - (aes_crypt_session().m_session_time + aes_crypt_session().m_session_timeout)) > 9 )
        {
            aes_crypt_session().m_session_time = 0; // Session abgelaufen.
            aes_crypt_session().init_crypt_mode();
            trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, aes_crypt_session().m_crypt_mode, (aes_crypt_session().m_session_time == 0) ? "no" : "yes" );
        }

        // Buffer fuer Folgebefehl zwischenspeichern,
        // da das hinter buf liegende Bytearray durch Aufruf von exchangeSessionKey veraendert wird.
        uint16_t lCount = count;
        uint8_t data[IV_PACKET_LENGTH_DATA];
        memset( data, 0, sizeof( data ) );
        memcpy( data, buffer, count );

        // Aushandlung eines Session-Key durchfuehren
        int lRes = exchange_session_key( get_device_address_receiver() );
        if ( lRes == 0x0D00 || lRes == -IV_ERROR_NEED_USER_PASSWORD )
            return -IV_ERROR_NEED_USER_PASSWORD;

        // Aushandlung erfolgreich durchgefuehrt, Parameterdaten wiederherstellen.
        count = lCount;
        memcpy( buffer, data, count );
    }

    int res = Protocol_base::call( context, module, func, buffer, count, context_value );
    // PROT_MODULE_ERROR_DECRYPT                (0x0D00) // Entschluesselung des Packetinhaltes bring Fehler.
    //                                                      Ablauf des Session-Key, Neuaushandlung ist erforderlich.
    if ( res == 0x0D00 )
    {
        //if ( module == IV_MODULE_PROTOCOL_CRYPT && func == IV_FUNCTION_EXCHANGE_PUBLIC_KEY )
        //  return res;

        aes_crypt_session().m_session_time = 0; // Session-Key ungueltig.
        aes_crypt_session().init_crypt_mode();
        trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %u, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, aes_crypt_session().m_crypt_mode, (aes_crypt_session().m_session_time == 0) ? "no" : "yes" );
        if ( repeat == 0 )
        {
            repeat++;
            goto label_again;
        }
        else
        {
            // Das aktuell angegebene User-Passwort ist ungueltig.
            return -IV_ERROR_NEED_USER_PASSWORD;
        }
    }

    // PROT_MODULE_ERROR_CRYPTED                (0x0E00) // Mode passt nicht. Z. B. Geraet arbeitet verschluesselt und Anfrage war unverschluesselt.
    if ( res == 0x0E00 )
    {
        // Fehler bei Entschluesselung der Daten.
        aes_crypt_session().m_crypt_mode = 0;
        aes_crypt_session().m_session_time = 0;
    }

    return res;
}

int Datafox::Communication::Protocol_crypt::write_protocol_crypt()
{
    // Handelt es sich um die vorhergehende Packetnummer?
    bool resend = (m_packet_number_previous == m_packet_send[IV_PACKET_LENGTH_HEAD]);
    // Packettyp uebernehmen
    //  unsigned char type = _packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3]];
    // Anzahl der Datenbyte
    uint16_t count = *(uint16_t *)&m_packet_send[4] - IV_PACKET_LENGTH_HEAD - m_packet_send[3] - sizeof( uint8_t ) - sizeof( uint16_t ) * 2;

    // Aktuell intern verwendeten Crypt-Mode eintragen.
    m_packet_send[6] |= aes_crypt_session().m_crypt_mode;

    // Wenn cryptMode > 1 (dann wird die Verschluesselung von der Gegenseite unterstuetzt),
    // dann verschluesseln und Verschluesselte Daten fuer ggf. wiederholtes
    // senden halten.
    if ( aes_crypt_session().m_crypt_mode > 1 )
    {
        uint16_t size;
        // Daten werden verschluesselt.
        m_packet_send[6] |= IV_PACKET_FLAG_CRYPTED;

        // Wiederholtes Senden?
        if ( resend )
        {
            // Zwischengespeicherten Dateninhalt wiederholt senden.
            memcpy( &m_packet_send[IV_PACKET_LENGTH_HEAD + m_packet_send[3] + sizeof( uint8_t )], m_packet_data_previous, m_packet_data_size_previous );
            size = m_packet_data_size_previous;
        }
        else
        {
            size = aes_crypt_session().send_encrypt( &m_packet_send[IV_PACKET_LENGTH_HEAD + m_packet_send[3] + sizeof( uint8_t )], sizeof( uint16_t ) * 2 + count );
            // Aktuelle Daten fuer ggf. wiederholtes Senden halten.
            memcpy( m_packet_data_previous, &m_packet_send[IV_PACKET_LENGTH_HEAD + m_packet_send[3] + sizeof( uint8_t )], size );
            m_packet_data_size_previous = size;
        }

        // Anzahl Bytes die fuer die Verschluesselung hinzugefuegt wurde (16Byte Bloecke) berechnen.
        size -= sizeof( uint16_t ) * 2 + count;
        // Auf die Gesamtgroesse des Pakets draufaddieren.
        *(uint16_t *)&m_packet_send[4] += size;
        m_packet_number_previous = m_packet_send[IV_PACKET_LENGTH_HEAD];
    }

    return IV_ERROR_SUCCESS;
}

int Datafox::Communication::Protocol_crypt::write_stream_crypt( std::string &buffer )
{
    // Wenn cryptMode > 1 (dann wird die Verschluesselung von der Gegenseite unterstuetzt),
    // dann verschluesseln und Verschluesselte Daten fuer ggf. wiederholtes
    // senden halten.
    if ( aes_crypt_session().m_crypt_mode > 1 )
    {
        uint32_t count_crypt_byte;
        uint32_t count_blocks;

        // Anzahl der Datenbyte
        uint32_t count = (uint32_t)buffer.size();

        // Anzahl 16 Byte Bloecke berechnen und Streamlaenge entsprechend anpassen.
        // Zusaetzlich werden die Daten um einen CRC und Anzahl Nutzdaten erweitert, daher (count + 5)!
        count_blocks = (count + 5) / 16;
        if ((count + 5) % 16)
            count_blocks++;

        count_crypt_byte = count_blocks * 16;

        // Puffer auf benoetigte Laenge anpassen.
        // Die Routine fuer Crypt fuellt den letzen 16 Byte Block des Puffer ggf. mit Zufallswerten auf.
        buffer.resize( count_crypt_byte, ' ' );

        // Datenblock verschluesseln.
        aes_crypt_session().send_encrypt_stream( (uint8_t *)&buffer[0], count );
    }

    return IV_ERROR_SUCCESS;
}

int Datafox::Communication::Protocol_crypt::read_protocol_crypt()
{
    // Packettyp uebernehmen
    //  unsigned char type = _packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3]];
    // Anzahl der Datenbyte
    uint16_t count = *(uint16_t *)&m_packet_receive[4] - IV_PACKET_LENGTH_FRAME - m_packet_receive[3] - sizeof( uint8_t ) - sizeof( uint16_t ) * 2;

    // Wenn cryptMode > 1 (dann wird die Verschluesselung unterstuetzt), 
    // zusaetzlich die Daten verschluesselt vorliegen, dann entschluesseln.
    if ( (aes_crypt_session().m_crypt_mode > 1) && (m_last_flags & IV_PACKET_FLAG_CRYPTED) )
    {
        uint16_t size = 2 * sizeof( uint16_t ) + count;
        if ( !aes_crypt_session().receive_decrypt( &m_packet_receive[IV_PACKET_LENGTH_HEAD + m_packet_receive[3] + sizeof( uint8_t )], &size ) )
            return 0x0D00;

        *(uint16_t *)&m_packet_receive[4] -= count - (size - 2 * sizeof( uint16_t ));
        count = size - 2 * sizeof( uint16_t );
    }

    // Liegt noch keine Information ueber die unterstuetzte Verschluesselung der Gegenstelle vor,
    // muss diese ermittelt werden.
    if ( aes_crypt_session().m_crypt_mode != (m_last_flags & 0x3) )
    {
        trace( "%s(%i):WARNING(1):  %s() [ cryptMode change %02X -> %02X ]\n", __FILE__, __LINE__, __func__, aes_crypt_session().m_crypt_mode, (m_last_flags & 0x3) );

        // Die verschluesselte Kommunikation wird auf eine der folgenden Arten unterstuetzt.
        if ( (m_last_flags & 0x3) == IV_PACKET_CRYPT_MODE_NO
            || (m_last_flags & 0x3) == IV_PACKET_CRYPT_MODE_DEFAULT
            || (m_last_flags & 0x3) == IV_PACKET_CRYPT_MODE_PASSPHRASE )
        {
            // Uebernahme des Verschluesselungsmode
            aes_crypt_session().m_crypt_mode = (m_last_flags & 0x3);
        }
        else
        {
            // Die verschluesselte Kommunikation wird nicht unterstuetzt,
            // der Verschluesselungsmode ist somit 0.
            aes_crypt_session().m_crypt_mode = 0;
        }

        // Initialisieren des aktuell angenommenen Crypt-Mode.
        aes_crypt_session().m_session_time = 0; // Eventuell aktueller Session-Key ist ungueltig.
        aes_crypt_session().init_crypt_mode();
        trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, aes_crypt_session().m_crypt_mode, (aes_crypt_session().m_session_time == 0) ? "no" : "yes" );
    }

    return IV_ERROR_SUCCESS;
}

int Datafox::Communication::Protocol_crypt::read_stream_crypt( std::string &buffer )
{
    // Wenn cryptMode > 1 (dann wird die Verschluesselung unterstuetzt), 
    // zusaetzlich die Daten verschluesselt vorliegen, dann entschluesseln.
    if ( aes_crypt_session().m_crypt_mode > 1 )
    {
        uint32_t size = (uint16_t)buffer.size();
        if ( !aes_crypt_session().receive_decrypt_stream( (uint8_t *)&buffer[0], &size ) )
            return 0x0D00;

        buffer.resize( size );
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Austausch eines Session-Key fuer die verschluesselte Kommunikation.
*
* @param deviceAddress Anzusprechende Geraetenummer.
*/
int Datafox::Communication::Protocol_crypt::exchange_session_key( uint8_t device_address )
{
    trace( "%s(%i):INFO(3):  %s( %i )\n", __FILE__, __LINE__, __func__, device_address );

    int result = -IV_ERROR_FAIL;
    uint16_t count = 0;
    uint8_t packet_data[IV_PACKET_LENGTH_DATA];

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Aufbau eines 8 Byte zufallsteils vom Session-Key
        uint8_t myKey[8];
        memset( myKey, 0, sizeof( myKey ) );
        uint8_t i;
        // Einen 8 Byte Langen Zufallswert erstellen.
        for ( i = 0; i < sizeof( myKey ); ++i )
            myKey[i] = rand() % 255;

        // Daten in Sendebuffer kopieren
        memcpy( packet_data, myKey, sizeof( myKey ) );
        count = sizeof( myKey );
        if ( aes_crypt_session().m_crypt_mode == IV_PACKET_CRYPT_MODE_PASSPHRASE )
            *(uint32_t *)&packet_data[count] = aes_crypt_session().m_session_timeout;
        else
            *(uint32_t *)&packet_data[count] = 3600; // 1 Stunde default
        
        count += 4;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_PROTOCOL_CRYPT, IV_FUNCTION_EXCHANGE_PUBLIC_KEY, packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace( "%s(%i):INFO(2): %s() [ %04X ]\n", __FILE__, __LINE__, __func__, result & 0xFFFF );
            return result;
        }

        // Antwortparameter pruefen
        if ( count != 8 )
        {
            trace( "%s(%i):WARNING(1):  %s() [ Antwortparameter fehlt ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Aktuellen Aushandlungszeitpunkt halten
        aes_crypt_session().m_session_time = time( NULL );

        // Gesamten Session-Key uebernehmen.
        memcpy( aes_crypt_session().m_session_key, myKey, 8 );
        // Uebernehmen des Session-Key Teils vom Geraet.
        memcpy( &aes_crypt_session().m_session_key[8], packet_data, 8 );
        aes_crypt_session().m_session_key_size = 16;

        // Schluesselaushandlung erfolgreich.
        trace( "%s(%i):INFO(1): %s() [ Session-Key valid ]\n", __FILE__, __LINE__, __func__ );

        // Initialisieren der Verschluesselung (verwendet m_sessionTime, m_sessionKey)
        aes_crypt_session().init_crypt_mode();
        trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, aes_crypt_session().m_crypt_mode, (aes_crypt_session().m_session_time == 0) ? "no" : "yes" );
    }

    return result;
}

int Datafox::Communication::Protocol_crypt::set_communication_password( const char* password, uint8_t size, uint32_t session_timeout, bool global ) 
{
    return aes_crypt_session().set_communication_password( password, size, session_timeout, global );
}
