/**
* @file  protocol_docking_station_v1.cpp
* @brief Implementierung der Klasse TimeboyStation.
*/
//////////////////////////////////////////////////////////////////////

#include "datafox/communication/protocol_docking_station_v1.h"
#include <memory.h>

// ##################################################################
/**
* @brief Destruktor schliesst die eventuell noch geoeffnete Schnittstelle. 
*
*/
Datafox::Communication::Protocol_dockingstation_v1::~Protocol_dockingstation_v1() 
{
}


//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

//#########################################################
//######################### CRC ###########################
//#########################################################

// ##################################################################
/**
* @brief Erzeugt einen 16bit CRC Wert fuer die uebergebenen Daten.
*/
unsigned short Datafox::Communication::Protocol_dockingstation_v1::get_crc16( unsigned char* data, unsigned int size, unsigned short crc )
{
    get_connection().trace( "%s(%i):INFO(3): %s [ 0x%08X, %i, 0x%04X ]\n", __FILE__, __LINE__, __func__, data, size, crc );

    unsigned int i;
    unsigned char j;

    for ( i = 0; i < size; ++i )
    {
        crc = crc ^ data[i];

        for ( j = 0; j < 8; ++j )
        {
            if ( crc & 0x1 )
            {
                crc >>= 1; 
                crc = crc ^ 0xA001;
            }
            else 
            {
                crc >>= 1;
            }
        }
    }

    return crc;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
unsigned short Datafox::Communication::Protocol_dockingstation_v1::read_slot_firmware_version( unsigned char device_address )
{
    unsigned short ver;
    int res;
    int autodetect = 1;

    // 254 kann nur Einzeldocking sein und nutzt MasterIV Protokoll.
    // Bereich pruefen
    if ( device_address < 1 || device_address >= 254 )
    {
        return 0;
    }

    switch ( m_timeboy_slot[device_address].get_version() ) 
    {
    case 0xFFFF: // Pruefung bereits durchgefuehrt und war nicht erfolgreich.
        break;

    case 0: 
        { // Ermitteln der FWVersionProtokollversion
            Datafox::Communication::Connection_base::Connection_type type;
            std::string commString;
            uint32_t commVal;
            uint32_t commTO;
            uint32_t commValOld = 0;
            get_connection().get_open_parameters( type, commString, commVal, commTO );

            while ( autodetect )
            {
                // Bei Seriell ist es wahrscheinlicher, dass mit 9600bd kommuniziert wird, deswegen zuerst damit Pruefen und dann mit 38400bd.
                if ( type == Datafox::Communication::Connection_base::Connection_type::serial && commVal != 9600 )
                {
                    // Schnittstellenbaud anpassen
                    commValOld = commVal;
                    commVal = 9600;
                    get_connection().change_state_serial( get_connection().get_serial_port(), 9600, commTO );
                }
                else 
                {
                    // Wurde Baud per autodetect abgeaendert?
                    if (commValOld)
                    {
                        get_connection().change_state_serial( get_connection().get_serial_port(), commValOld, commTO );
                    }

                    autodetect = 0;
                }

                // In der Version 1.25 (08.2007) wurde die Timeboyerkennung per Kommunikationspolling der Docking auf TB vorgenommen.
                // Ab 1.26 (09.2007) wurde die Unterstützung der Timeboyerkennung per Komperatorschaltung integriert.
                // De-/Aktivierung über 0x8000000C (Wert 2 kann dort verwendet werden), abfrage des Zustandes über 0x8000E000.

                if ( (res = read_bytes( device_address, MB_SLOT_VERSION, 0, (unsigned char*)&ver, sizeof( ver ) )) == IV_ERROR_SUCCESS )
                {
                    // < 1.31	Paketpuffer ist 45 Byte
                    // >= 1.31	Paketpuffer ist 150 Byte
                    m_timeboy_slot[0].set_version( commVal ); // Im 0 Byte die verwendete Baudrate halten.
                    m_timeboy_slot[device_address].set_version( bin_to_uint_msb_first( (unsigned char*)&ver ) );
                    return m_timeboy_slot[device_address].get_version();
                }
                else
                {
                    m_timeboy_slot[device_address].set_version( 0xFFFF );
                }
            }
            break;
        }

    default:
        return m_timeboy_slot[device_address].get_version();
    }

    return 0;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Datafox::Communication::Protocol_dockingstation_v1::reset_slot( unsigned char device_address )
{
    if ( device_address < 0 || device_address >= 254 )
    {
        return;
    }

    m_timeboy_slot[device_address].reset();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Datafox::Communication::Protocol_dockingstation_v1::reset_slots()
{
    unsigned char device_address;
    for ( device_address = 0; device_address < 255; ++device_address )
    {
        m_timeboy_slot[device_address].reset();
    }
}

int Datafox::Communication::Protocol_dockingstation_v1::call_docking( unsigned char* send_buffer, unsigned char send_count, unsigned char* receive_buffer, unsigned char* receive_count )
{
    int res;
    std::size_t nb;
    std::size_t rcnt = 0;
    std::size_t len;

    // Autobaud
    // Stimmt die aktuell Verwendete Baudrate nicht mit der benoetigten ueberein?
    if ( get_connection().get_type() == Datafox::Communication::Connection_base::Connection_type::serial
        && m_timeboy_slot[0].get_version() && get_connection().get_port_number() != m_timeboy_slot[0].get_version() )
    {
        get_connection().change_state_serial( get_connection().get_serial_port(), m_timeboy_slot[0].get_version(), get_connection().get_read_timeout() );
    }

    // CRC noch an Paket anfuegen
    *(unsigned short*)&send_buffer[send_count] = get_crc16( send_buffer, send_count );
    send_count += sizeof( unsigned short );

    // Empfangsanzahl initialisieren
    *receive_count = 0;

#ifdef TRACE_DOCKING
    get_connection().trace( "%s(%i):INFO(1): %s.send( %i ) ( %s[ 0x%02X ], Add: 0x%08X  Cnt: %i )\n", __FILE__, __LINE__, __func__, send_buffer[0], (send_buffer[3] == MB_READ) ? "READ" : "WRITE", send_buffer[3], bin_to_ulong_msb_first(&send_buffer[4]), (send_buffer[3] == MB_READ) ? send_buffer[9] : send_buffer[2] - 0x0A );
#endif

    // Paketdaten versenden.
    if ( (res = get_connection().write_data( send_buffer, send_count, nb, 1 )) != IV_ERROR_SUCCESS )
    {
        get_connection().trace( "%s(%i):ERROR(1): %s.transmit [ TransmitData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
        // Neuerkennung notwendig.
        reset_slot( device_address_receiver() );
        return res;
    }
    // Alle Zeichen versendet?
    if ( nb != (int)send_count )
    {
        return -IV_ERROR_TRANSMIT;
    }

    // Kopf des Paketes empfangen.
    if ( (res = get_connection().read_data( receive_buffer, 4, nb, 1 )) != IV_ERROR_SUCCESS )
    {
        get_connection().trace( "%s(%i):ERROR(1): %s [ ReceiveData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
        // Neuerkennung notwendig.
        reset_slot( device_address_receiver() );
        return res;
    }
    // Alle Zeichen empfangen?
    if ( nb < 4 ) 
    {
        get_connection().trace( "%s(%i):ERROR(1): %s [ %i von %i Byte empfangen ]\n", __FILE__, __LINE__, __func__, nb, 4 );
        return -IV_ERROR_RECEIVE;
    }
    rcnt = nb;

    // Busadresse vergleichen
    if ( send_buffer[0] != receive_buffer[0] )
    {
        //#########################
        // Bugfix fuer Dockingstation FW Version 1.24
        // Naechstes Byte ist Adresse?
        if ( send_buffer[0] == receive_buffer[1] )
        {
            //getChannelBase().Trace("%s(%i):ERROR(1): %s [ Skip wrong start byte {0x%02X} ]\n", __FILE__, __LINE__, __func__, rxBuf[0]);
            receive_buffer[0] = receive_buffer[1];
            receive_buffer[1] = receive_buffer[2];
            receive_buffer[2] = receive_buffer[3];
            rcnt--;
        }
        else
        {
            // Delay bei TCP/IP ueber Com-Server
            if ( get_connection().get_type() == Datafox::Communication::Connection_base::Connection_type::socket )
            {
                std::this_thread::sleep_for( std::chrono::milliseconds( 200 ) );
            }

            get_connection().trace( "%s(%i):ERROR(1): %s [ address wrong %i != %i ]\n", __FILE__, __LINE__, __func__, send_buffer[0], receive_buffer[0] );
#ifdef _DEBUG
            int i; 
            for ( i = 0; i < nb; ++i ) 
            {
                get_connection().trace( "%02X ", receive_buffer[i] );
            }
            get_connection().trace( "\n" );
#endif
            get_connection().flush();
            return -IV_ERROR_ADDRESS;
        }
    }

    // Solllaenge Paket aus Kopf ermitteln
    len = bin_to_uint_msb_first( &receive_buffer[1] );
    if ( len < 7 || len > 200 ) 
    {
        // Delay bei TCP/IP ueber Com-Server
        if ( get_connection().get_type() == Datafox::Communication::Connection_base::Connection_type::socket )
        {
            std::this_thread::sleep_for( std::chrono::milliseconds( 200 ) );
        }

        get_connection().trace( "%s(%i):ERROR(1): %s [ packet length < 7 || > 200 {%i}]\n", __FILE__, __LINE__, __func__, len );
#ifdef _DEBUG
        int i; 
        for ( i = 0; i < nb; ++i ) 
        {
            get_connection().trace( "%02X ", receive_buffer[i] );
        }
        get_connection().trace( "\n" );
#endif
        get_connection().flush();
        return -IV_ERROR_PACKETLENGTH;
    }

    // Body des Paketes empfangen.
    if ( (res = get_connection().read_data( &receive_buffer[rcnt], len-rcnt, nb, 1 )) != IV_ERROR_SUCCESS ) 
    {
        get_connection().trace( "%s(%i):ERROR(1): %s [ ReceiveData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
        // Neuerkennung notwendig.
        reset_slot( device_address_receiver() );
        return res;
    }
    // Alle Zeichen empfangen?
    if ( nb < (len - rcnt) ) 
    {
        get_connection().trace( "%s(%i):ERROR(1): %s [ Nachlesen %i von %i Byte empfangen ]\n", __FILE__, __LINE__, __func__, nb, (len - rcnt) );
        get_connection().flush();
        return -IV_ERROR_RECEIVE;
    }
    rcnt += nb;

    // Alle Zeichen empfangen?
    if ( get_crc16( receive_buffer, rcnt ) ) 
    {
        get_connection().trace( "%s(%i):ERROR(1): %s [ Paket CRC ]\n", __FILE__, __LINE__, __func__ );
        return -IV_ERROR_CRC_PACKET;
    }

    // Handelt es sich um eine Steuerantwort der Dockingstation?
    if ( rcnt == 7 )
    { 
#ifdef TRACE_DOCKING
        get_connection().trace( "%s(%i):INFO(1): %s.receive ( Befehl: 0x%02X, Fehler: 0x%02X )\n", __FILE__, __LINE__, __func__, receive_buffer[3], receive_buffer[4] );
#endif
    }
    else
    {
#ifdef TRACE_DOCKING
        get_connection().trace( "%s(%i):INFO(1): %s.receive ( Befehl: 0x%02X, Länge: %i )\n", __FILE__, __LINE__, __func__, receive_buffer[3], rcnt );
#endif
    }

    // Zeitstempel der Antwort halten
    m_timeboy_slot[device_address_receiver()].set_receive_stamp_slot( clock() );

    // Anzahl empfangener Bytes
    *receive_count = rcnt;

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Prueft die Erreichbarkeit des Kommunikationspartners.
*
* @param deviceID Anzusprechender Kommunikationspartner.
* @param model Erreichtes Timeboymodel.
*/
int Datafox::Communication::Protocol_dockingstation_v1::read_timeboy_model( unsigned char device_address, int& model )
{
    get_connection().trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );
    int res = IV_ERROR_SUCCESS;
    unsigned char data[2];
    memset( data, 0, sizeof( data ) );

    // Ist die schnelle Steck-Erkennung verfuegbar?
    if ( m_timeboy_slot[device_address].get_detection_state() == 2 )
    {
        // Pruefung ob der Slot auch meldet, dass ein Timeboy steckt.
        if ( (res = read_bytes( device_address, MB_SLOT_STATE, 0, data, 2, 1 )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            return res;
        }
        if ( *(unsigned short*)data != 0x0100 )
        {
            // Timeboy ist nicht erreichbar, da nicht gesteckt!
            //			return -IV_ERROR_UNREACHABLE;
            // KORREKTUR (2012_05_14):
            // Da ab der TimeboyIV Version 04.02.02.40 die Ladeabschaltung voll unterstuetzt wird,
            // Kommt es durch die Abfrageroutine zu Fehlinformationen, da die Ladeschaltung meldet,
            // das kein Timeboy steckt obwohl dieses der Fall ist.
        }
    }

    do 
    {
        // Timeboy ansprechen.
        if ( (res = read_bytes( device_address, 0, 0, data, 2, 1 )) != IV_ERROR_SUCCESS )
        {
            // Wurde ermittelt, dass ein Timeboy steckt? -> Nochmal versuchen.
            if ( *(unsigned short*)data == 0x0100 )
            {
                *(unsigned short*)data = 0;
                continue;
            }

            // Fehler ausgeben
            return res;
        }
    } while ( res != IV_ERROR_SUCCESS );

    if ( data[0] == '~' && data[1] == '~' )
    {
        model = 4;
    }
    else
    {
        model = 3;
    }

    // Unabhaengig vom Timeboymodel
    // Pruefung auf die Verfuegbarkeit der Timeboyerkennung durchfuehren.
    if ( m_timeboy_slot[device_address].get_version() >= 0x0126 && m_timeboy_slot[device_address].get_detection_state() == 0 )
    {
        if ( (res = read_bytes( device_address, MB_SLOT_CHECK, 0, data, 2, 1 )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            return res;
        }

        if ( *(unsigned short*)data == 0x0002 )
        {
            // Wird unterstuetzt
            m_timeboy_slot[device_address].set_detection_state( 2 );
        }
        else if ( *(unsigned short*)data == 0x0001 )
        {
            // Pruefung ob der Slot auch meldet, dass ein Timeboy steckt.
            if ( (res = read_bytes( device_address, MB_SLOT_STATE, 0, data, 2, 1 )) != IV_ERROR_SUCCESS )
            {
                // Fehler ausgeben
                return res;
            }

            if ( *(unsigned short*)data == 0x0100 )
            {
                // Timeboy wurde erkannt und der Slot meldet, dass ein Timeboy steckt -> Komperatorschaltung bestueckt.
                *(unsigned short*)data = 0x0002;
                //if ((m_result = writeBytes(deviceID, MB_SLOT_CHECK, data, 2, 1)) != IV_ERROR_SUCCESS)
                //{
                //	// Fehler ausgeben
                //	return m_result;
                //}

                m_timeboy_slot[device_address].set_detection_state( 2 );
            }
            else
            {
                m_timeboy_slot[device_address].set_detection_state( 1 );
            }
        }
        else
        {
            m_timeboy_slot[device_address].set_detection_state( 1 );
        }
    }

    return res;
}

// ##################################################################
/**
* @brief Setzt den Zustand der Roten LED der Mehrfachdocking.
*
* @param deviceID Anzusprechender Kommunikationspartner.
*/
int Datafox::Communication::Protocol_dockingstation_v1::set_docking_led( unsigned char device_address, int state )
{
    get_connection().trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );
    int res = IV_ERROR_SUCCESS;
    unsigned char data[1];

    if ( state != 0 )
    {
        data[0] = 0x00;
    }
    else
    {
        data[0] = 0xFF;
    }

    if ( (res = write_bytes( device_address, MB_SLOT_LED, 0, data, 1 )) != IV_ERROR_SUCCESS )
    {
        // Fehler ausgeben
        return res;
    }

#ifdef TRACE_DOCKING
    get_connection().trace( "%s(%i):INFO(1): %s [ Letzter Wechsel vor %0.3f Sekunden ]\n", __FILE__, __LINE__, __func__, float(clock() - m_timeboy_slot[device_address].get_led_stamp()) / float(CLOCKS_PER_SEC) );
#endif

    // Zustand der LED und Zeitpunkt des wechsels halten
    m_timeboy_slot[device_address].set_led_state( (state != 0) ? 0x01 : 0xFF );
    m_timeboy_slot[device_address].set_led_stamp( clock() );

    return res;
}

// ##################################################################
/**
* @brief Setzt den Kommunikationszustand eines Slots der Mehrfachdocking.
*
* @param deviceID Anzusprechender Kommunikationspartner.
* @param state Zu setzender Kommunikationszustand.
*/
int Datafox::Communication::Protocol_dockingstation_v1::set_timeboy_slot_state( unsigned char device_address, int state, int command_result )
{
    (void)command_result;
    get_connection().trace( "%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, state );

    if ( is_slot( device_address ) )
    {
        // Ist der geforderte Zustand gleich dem aktuellen?
        if  (m_timeboy_slot[device_address].get_led_state() == state )
        {
            return IV_ERROR_SUCCESS;
        }

        if ( state != 0 )
        {
            int model = 0;
            // Es muss ein gueltiges Timeboymodel vorliegen
            if ( read_timeboy_model( device_address, model ) != IV_ERROR_SUCCESS )
            {
                return -IV_ERROR_UNREACHABLE;
            }

            if ( model != 4 )
            {
                return -IV_ERROR_TIMEBOY_I_II_III;
            }

            // Rote LED einschalten
            if ( set_docking_led( device_address, 1 ) != IV_ERROR_SUCCESS )
            {
                return -IV_ERROR_UNREACHABLE;
            }
        }
        else
        {
            // Rote LED ausschalten
            if( set_docking_led( device_address, 0 ) != IV_ERROR_SUCCESS )
            {
                return -IV_ERROR_UNREACHABLE;
            }
        }
    }
    else
    {
        // Bei der angesprochenen ID handelt es sich um keinen Slot einer TimeboyStation.
        return -IV_ERROR_SLOT_NONE;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Entgegennehmen der Verbindung und starten eines Verarbeitungsthreads.
*/
int Datafox::Communication::Protocol_dockingstation_v1::start_communication_thread() 
{
    get_connection().trace( "%s(%i):INFO(3): %s [  ]\n", __FILE__, __LINE__, __func__ );

    // Hauptthread starten
    if ( m_communication_thread.joinable() == false )
    {
        m_communication_thread = std::thread( [=] { communication_thread(); } );

        int i;
        // Hochlauf des Workers pruefen
        for ( i = 0; m_communication_thread.joinable() && (get_thread_state() != ts_idle) && (i < 60); i++ )
        {
            std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );
        }
    }

    return m_communication_thread.joinable();
}

// ##################################################################
/**
* @brief Beendigungssignal im Workerthread setzen und auf dessen beendigung warten.
*/
void Datafox::Communication::Protocol_dockingstation_v1::stop_communication_thread()
{
    get_connection().trace( "%s(%i):INFO(3): %s", __FILE__, __LINE__, __func__ );

    // Socket-Thread deaktivieren, wenn aktiv
    if ( m_communication_thread.joinable() )
    {
        // WorkerThread fuer Socket beenden.
        set_thread_command( tc_exit );

        // Thread Beendigung abwarten (communicationThread stellt m_endSignal auf FALSE)
        m_communication_thread.join();
    }
}

// ##################################################################
/**
* @brief Workerthread fuer die Verarbeitung der Kommunikationsaufgaben.
*/
void Datafox::Communication::Protocol_dockingstation_v1::communication_thread()
{
    get_connection().trace( "%s(%i):INFO(3): %s\n", __FILE__, __LINE__, __func__ );

    // Threadzustand setzen
    m_thread_state = ts_begin;
    m_thread_command = tc_none;

    // Hauptschleife ausfuehren?
    bool run = true;

    // Hauptschleife des Worker
    while ( run )
    {
        switch ( m_thread_state )
        {
        case ts_begin:
            // Threadzustand setzen
            m_thread_state = ts_idle;
            break;

        case ts_idle: {
            if ( m_thread_command != tc_none )
            {
                // Threadzustand setzen
                m_thread_state = ts_execute_command;
            }
            else
            {
                //// Pruefung ob LED augeschaltet werden muss.
                //if (IsFlag(TSF_AUTO_LED))
                //{
                //	for (deviceID=1; deviceID < 255; deviceID++)
                //	{
                //		if (_timeboySlot[deviceID].getLEDState() && ((clock() - _timeboySlot[deviceID].getLEDStamp()) > 3))
                //		{
                //			setDockingLED(deviceID, 0);
                //		}
                //	}
                //}

                std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );
            }
            break;}

        case ts_execute_command: {

            switch ( m_thread_command )
            {
            case tc_none:
                break;

            case tc_exit:
                // Hauptschleife verlassen
                run = false;
                break;

            case tc_write:
                m_command_parameter.res = thread_write_bytes( m_command_parameter.device_address_receiver, m_command_parameter.base_address, m_command_parameter.offset, m_command_parameter.buffer, m_command_parameter.count, m_command_parameter.trials );
                break;

            case tc_read:
                m_command_parameter.res = thread_read_bytes( m_command_parameter.device_address_receiver, m_command_parameter.base_address, m_command_parameter.offset, m_command_parameter.buffer, m_command_parameter.count, m_command_parameter.trials );
                break;
            }

            // Aufgabe abgearbeitet
            m_thread_state = ts_idle;
            m_thread_command = tc_none;
            break;}

                                //case none:
        case ts_end:
        case ts_dummy:
            break;
        }
    }

    // Threadzustand setzen
    m_thread_state = ts_end;

    get_connection().trace( "%s(%i):INFO(1): %s [ End ]\n", __FILE__, __LINE__, __func__ );
}

//#########################################################
//################## Members ##############################
//#########################################################

/*############################################################################*/
//  FUNKTION:    writeBytes
/** Schreibbefehl, um Bytes in den Timeboy zu schreiben.

Autor                 : Michael Wicher            @n
erstellt am           : 02.03.2007                @n

@param[in] address Zieladresse im Timeboy.
@param[in] buf Zeiger auf Daten, die gesendet werden sollen.
@param[in] count Anzahl Bytes zum Senden.
@param[in] slot Steckplatz der Docking [1..10].

@return #TRUE -> Erfolg, #FALSE -> Miszerfolg.

@b Aufgaben:
- Bytes an den Timeboy senden.

@b Aenderungen:
- 17.01.2008 (MW) Neuer Transferbefehl, um die Kommunikation durch einen Mutex
zu schützen.
- 01.02.2008 (MW) Bei allen Befehlen für den Timeboy wird zuvor bei der Docking
angefragt, ob auch ein Timeboy steckt.
Weiterhin wird hier geprüft, ob sich der Task terminieren soll.
Falls kein Timeboy steckt, wird hier auch die Startphase
abgeschloszen, so dasz die Funktion #DOCKINGControl beendet wird.
- 27.03.2008 (MW) Max. fünf Versuche, den Befehl abzusetzen. Dazwischen wird
immer geprüft, ob der Timeboy laut Docking steckt.
- 19.05.2008 (MW) Vor Abarbeitung des Befehls wird geprüft, ob der ausgewählte
Steckplatz der Docking noch erreichbar ist.
Bei Befehlen dirket für die Docking werden nun 2 Versuche
durchgeführt.
Die CRC-Berechnung verschoben.
*/
/*############################################################################*/

int Datafox::Communication::Protocol_dockingstation_v1::write_bytes( unsigned char device_address, unsigned int base_address, int offset, unsigned char* buffer, int count, int write_trials )
{
    int res = IV_ERROR_SUCCESS;

    // Ueber den Kommunikationsthread kommunizieren?
    if ( m_communication_thread.joinable() )
    {
        m_command_parameter.device_address_receiver = device_address;
        m_command_parameter.base_address = base_address;
        m_command_parameter.offset = offset;
        m_command_parameter.buffer = buffer;
        m_command_parameter.count = count;
        m_command_parameter.trials = write_trials;

        // Schreibvorgang starten.
        if ( set_thread_command( tc_write ) )
        {
            while ( get_thread_command() != tc_none )
            {
                std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );
            }
        }
        else
        {
            return -IV_ERROR_MODULE_BUSY;
        }

        return m_command_parameter.res;
    }

    // Direkte Befehlsausfuehrung
    res = thread_write_bytes( device_address, base_address, offset, buffer, count, write_trials );

    if ( res == IV_ERROR_SUCCESS )
    {
        // Nach erfolgreicher Kommunikation automatisch den
        // Kommunikationsthread starten, wenn noch nicht geschehen.
        //		startCommunicationThread();
    }

    return res;
}

int Datafox::Communication::Protocol_dockingstation_v1::thread_write_bytes( unsigned char device_address, unsigned int base_address, int offset, unsigned char* buffer, int count, int write_trials )
{
    static unsigned char pre[MB_SLOT_BUFFER_SIZE_LARGE];

    int res = IV_ERROR_SUCCESS;
    unsigned char current[MB_SLOT_BUFFER_SIZE_LARGE];
    unsigned char i;
    unsigned char limit;
    unsigned char send[MB_SLOT_BUFFER_SIZE_LARGE];
    unsigned short pause;
    unsigned char receive[MB_SLOT_BUFFER_SIZE_LARGE];
    unsigned char repeat;
    unsigned char receive_count;
    unsigned char sendPre;
    int executeFails = 0;
    unsigned int address = base_address + offset;

    // Remoteadresse setzen.
    set_device_address_receiver( device_address );

    unsigned char bufSize = get_data_buffer_size( device_address_receiver() );

    // prüfe, ob die Busnummer verstellt ist
    //DOCKINGBusnumVerify(slot);

    for ( ; count; ) 
    {
        sendPre = FALSE;

        for ( limit = 20; limit; limit >>= 1 ) 
        {
            // bei allen Befehlen für den Timeboy bei der Docking anfragen, dasz auch ein Timeboy steckt
            //if ((((UINT8 *)&address)[3] & 0xF0) == 0x40) {

            //	KERNELCheckForTerminate(DOCKINGStop);

            //	// falls kein Timeboy steckt, macht der weitere Ablauf keinen Sinn und man kann abbrechen
            //	if (DOCKINGCheckTimeboy(slot) != DOCKING_TIMEBOY_PUSHED) {

            //		// Startphase abgeschloszen, jetzt darf auch der IO-Task loslaufen
            //		globalDOCKING.ready = TRUE;

            //		return (FALSE);

            //	}

            //} else {

            //	limit = 1;

            //}

            current[0] = (unsigned char)device_address_receiver();
            current[1] = 0x00;
            current[2] = 0x0A + ((count > bufSize) ? bufSize : count);
            current[3] = 0x22;
            current[4] = ((unsigned char *)&address)[3];
            current[5] = ((unsigned char *)&address)[2];
            current[6] = ((unsigned char *)&address)[1];
            current[7] = ((unsigned char *)&address)[0];

            memcpy( (unsigned char *)&current[8], buffer, (count > bufSize) ? bufSize : count );

            repeat = limit;

            // je nach Timeboy oder Docking als Empfänger
            if ( current[4] == 0x80 ) 
            {
                // beim Stellen der Busnummer nur ein Versuch
                if ( (current[6] == 0x00) && (current[7] == 0x00) ) 
                {
                    repeat = 1;
                    pause = 0;

                    // bei allen anderen Anfragen an die Docking zwei Versuche!
                } 
                else 
                {
                    repeat = 2;
                    pause = 80;
                }
            } 
            else 
            {
                pause = 2000 / repeat;
            }

            for ( i = 0; i < repeat; ++i ) 
            {
                //if (i) DEBUGString(TRACE_DEVICE_DOCKING, "repeat write <%u>.", i);

                if ( sendPre == TRUE ) 
                {
                    // lade, falls nötig,  den Vorgänger
                    memcpy( send, pre, sizeof( send ) );
                } 
                else 
                {
                    // aktuelles Paket laden
                    memcpy( send, current, sizeof( send ) );
                }

                if ( (res = call_docking( send, send[2] - 2, receive, &receive_count )) == IV_ERROR_SUCCESS ) 
                {
                    // Anzahl fehlversuche zurücksetzen
                    executeFails = 0;

                    // Paket für Docking?
                    if ( send[4] == 0x80 ) 
                    {
                        if ( *(unsigned int *)&receive[4] == *(unsigned int *)&send[4] ) 
                        {
                            break;
                        }
                    } 
                    else // Paket fuer Timeboy
                    {
                        if ( receive_count == 7 ) 
                        {
                            //DEBUGString(TRACE_DEVICE_DOCKING, "%02X, %08lX", receive[4], CONVSwapLong(*(UINT32 *)&transmit[4]));

                            if ( receive[4] == 0x22 ) 
                            {
                                //KERNELMilliSleep(100);
                                std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );

                                // falls nicht der Vorgänger gesendet wurde
                                if ( sendPre == FALSE ) 
                                {
                                    // gesendetes Paket zur evtl. Wiederholung merken
                                    memcpy( pre, send, sizeof( pre ) );

                                    // fertig
                                    break;
                                }

                                // Limit neu festlegen
                                repeat = limit;

                                // jetzt das aktuelle Paket nochmal senden
                                sendPre = FALSE;
                            } 
                            else if ( (receive[4] == 0x21) || (receive[4] == 0x24) ) 
                            {
                                //DEBUGString(TRACE_DEVICE_DOCKING, "repeat pre.");

                                // falls eine Wiederholung begonnen wird
                                if ( sendPre == FALSE ) 
                                {
                                    // lege Limit neu fest
                                    repeat = limit;
                                }

                                sendPre = TRUE;
                            }
                        }
                        else
                        {
                            //if (*(unsigned int *)&receive[4] == *(unsigned int *)&transmit[4]) 
                            //{
                            //	break;
                            //}
                        }
                    }
                }
                else
                {
                    // Fehlerhafte Ausfuehrungen halten
                    executeFails++;
                    if ( executeFails >= write_trials )
                    {
                        return res;
                    }
                }

                if ( pause ) 
                {
                    std::this_thread::sleep_for( std::chrono::milliseconds( pause ) );
                }
            }

            // bei erfolgreichem Senden keine Wiederholung
            if ( i < repeat ) 
            {
                break;
            }
        }

        // falls das Limit erreicht wurde, konnte der Befehl nicht erfolgreich gesendet werden!
        if ( limit == 0 ) 
        {
            return -IV_ERROR_UNREACHABLE;
        }

        address += bufSize;
        buffer += bufSize;
        count -= (count > bufSize ? bufSize : count);
    }

    // Zeitpunkt der letzten Timeboykommunikation halten
    if ( (address & 0x80000000) == 0 )
    {
        m_timeboy_slot[device_address_receiver()].set_receive_stamp_timeboy( clock() );
    }

    return res;
}

/*############################################################################*/
//  FUNKTION:    readBytes
/** Lesebefehl, um Bytes aus dem Timeboy zu lesen.

Autor                 : Michael Wicher            @n
erstellt am           : 02.03.2007                @n

@param[in] buf Zeiger, wo gelesene Daten hinkopiert werden sollen.
@param[in] address Quelladresse im Timeboy.
@param[in] count Anzahl Bytes zum Lesen.
@param[in] slot Steckplatz der Docking [1..10].

@return #TRUE -> Erfolg, #FALSE -> Miszerfolg.

@b Aufgaben:
- Bytes vom Timeboy lesen.

@b Aenderungen:
- 17.01.2008 (MW) Neuer Transferbefehl, um die Kommunikation durch einen Mutex
zu schützen.
- 01.02.2008 (MW) Bei allen Befehlen für den Timeboy wird zuvor bei der Docking
angefragt, ob auch ein Timeboy steckt.
Weiterhin wird hier geprüft, ob sich der Task terminieren soll.
Falls kein Timeboy steckt, wird hier auch die Startphase
abgeschloszen, so dasz die Funktion #DOCKINGControl beendet wird.
- 27.03.2008 (MW) Max. fünf Versuche, den Befehl abzusetzen. Dazwischen wird
immer geprüft, ob der Timeboy laut Docking steckt.
- 19.05.2008 (MW) Vor Abarbeitung des Befehls wird geprüft, ob der ausgewählte
Steckplatz der Docking noch erreichbar ist.
Bei Befehlen dirket für die Docking werden nun 2 Versuche
durchgeführt.
Die CRC-Berechnung verschoben.
*/
/*############################################################################*/

int Datafox::Communication::Protocol_dockingstation_v1::read_bytes( unsigned char device_address, unsigned int base_address, int offset, unsigned char* buffer, int count, int read_trials )
{
    int res = IV_ERROR_SUCCESS;
    // Ueber den Kommunikationsthread kommunizieren?
    if ( m_communication_thread.joinable() )
    {
        m_command_parameter.device_address_receiver = device_address;
        m_command_parameter.base_address = base_address;
        m_command_parameter.offset = offset;
        m_command_parameter.buffer = buffer;
        m_command_parameter.count = count;
        m_command_parameter.trials = read_trials;

        // Lesevorgang starten.
        if ( set_thread_command( tc_read ) )
        {
            while ( get_thread_command() != tc_none )
            {
                //connection().trace("%s(%i):INFO(1): %s [ wait ]\n", __FILE__, __LINE__, __func__);
                std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );
            }
        }
        else
        {
            return -IV_ERROR_MODULE_BUSY;
        }

        return m_command_parameter.res;
    }

    // Direkte Befehlsausfuehrung
    res = thread_read_bytes( device_address, base_address, offset, buffer, count, read_trials );

    if ( res == IV_ERROR_SUCCESS )
    {
        // Nach erfolgreicher Kommunikation automatisch den
        // Kommunikationsthread starten, wenn noch nicht geschehen.
        //		startCommunicationThread();
    }

    return res;
}

int Datafox::Communication::Protocol_dockingstation_v1::thread_read_bytes( unsigned char device_address, unsigned int base_address, int offset, unsigned char* buffer, int count, int read_trials )
{
    int res = IV_ERROR_SUCCESS;
    unsigned char i;
    unsigned char limit;
    unsigned char send[MB_SLOT_BUFFER_SIZE_LARGE];
    unsigned short pause;
    unsigned char receive[MB_SLOT_BUFFER_SIZE_LARGE];
    unsigned char repeat;
    unsigned char receive_count;
    int executeFails = 0;
    int error21Cnt = 0;
    unsigned int address = base_address + offset;

    // Remoteadresse setzen.
    set_device_address_receiver( device_address );

    unsigned char bufSize = get_data_buffer_size( device_address_receiver() );

    if ( offset == MB_SLOT_BUFFER_DATA_SIZE_SMALL )
    {
        address = base_address;
        if ( bufSize == MB_SLOT_BUFFER_DATA_SIZE_SMALL || (count - MB_SLOT_BUFFER_DATA_SIZE_SMALL) <= MB_SLOT_BUFFER_DATA_SIZE_SMALL )
        {
            // Nur wenn der moegliche Lesepuffer groesser ist, macht es sinn die bereits gelesenen Daten nochmals mitzulesen?
            // Die Dockingstation liest ein Packet vor und sendet immer die zuletzt angeforderte Packetgroesse in Folge.
            address += MB_SLOT_BUFFER_DATA_SIZE_SMALL;
            // bereits gelesene Daten von der geforderten Groesse abziehen
            count -= MB_SLOT_BUFFER_DATA_SIZE_SMALL;
            // Puffer weitersetzen
            buffer += MB_SLOT_BUFFER_DATA_SIZE_SMALL;
        }
    }

    // prüfe, ob die Busnummer verstellt ist
    //DOCKINGBusnumVerify(slot);

    for ( ; count; ) 
    {
        for ( limit = 20; limit; limit >>= 1 ) 
        {
            // bei allen Befehlen für den Timeboy bei der Docking anfragen, dasz auch ein Timeboy steckt
            //if ((((UINT8 *)&address)[3] & 0xF0) == 0x40) {

            //	KERNELCheckForTerminate(DOCKINGStop);

            //	// falls kein Timeboy steckt, macht der weitere Ablauf keinen Sinn und man kann abbrechen
            //	if (DOCKINGCheckTimeboy(slot) != DOCKING_TIMEBOY_PUSHED) {

            //		// Startphase abgeschloszen, jetzt darf auch der IO-Task loslaufen
            //		globalDOCKING.ready = TRUE;

            //		return (FALSE);

            //	}

            //} else {

            //	limit = 1;

            //}

            send[0] = (unsigned char)device_address_receiver();
            send[1] = 0x00;
            send[2] = 0x0C;
            send[3] = 0x21;
            send[4] = ((unsigned char *)&address)[3];
            send[5] = ((unsigned char *)&address)[2];
            send[6] = ((unsigned char *)&address)[1];
            send[7] = ((unsigned char *)&address)[0];
            send[8] = 0x00;
            send[9] = (count > bufSize) ? bufSize : count;

            repeat = limit;

            // je nach Timeboy oder Docking als Empfänger
            if ( send[4] == 0x80 ) 
            {
                repeat = 2;
                pause = 80;
            } 
            else 
            {
                pause = 2000 / repeat;
            }

            for ( i = 0; i < repeat; ++i ) 
            {
                //if (i) DEBUGString(TRACE_DEVICE_DOCKING, "repeat <%u>.", i);

                if ( (res = call_docking( send, 0x0A, receive, &receive_count )) == IV_ERROR_SUCCESS ) 
                {
                    // Fehlerhafte Ausfuehrungen zuruecksetzen
                    executeFails = 0;

                    if ( send[4] == 0x80 )
                    {
                        // wurde ein gültiges Antwortpaket empfangen, prüfen, ob auch die Länge paszt
                        if ( *(unsigned int *)&receive[4] == *(unsigned int *)&send[4] ) 
                        {
                            break;
                        }
                    }
                    else
                    {
                        if ( receive_count == 7 )
                        {
                            switch ( receive[4] )
                            {
                            case 0x22:
                            case 0x23:
                                break;
                            case 0x21:
                            case 0x24:
                                {
                                    // Erster Befehl kann aufgrund von Sleep unbeantwortet bleiben.
                                    error21Cnt++;
                                    if ( error21Cnt > 1 )
                                    {
                                        return -IV_ERROR_UNREACHABLE;
                                    }
                                }
                                break;
                            }
                        }
                        else
                        {
                            // Fehlerzaehler ruecksetzen.
                            error21Cnt = 0;

                            // wurde ein gueltiges Antwortpaket empfangen, pruefen, ob auch die Adresse passt
                            if ( *(unsigned int *)&receive[4] == *(unsigned int *)&send[4] ) 
                            {
                                break;
                            }
                        }
                    }
                }
                else
                {
                    // Fehlerhafte Ausfuehrung
                    executeFails++;
                    if ( executeFails >= read_trials )
                    {
                        return res;
                    }
                }

                //KERNELMilliSleep(pause);
                std::this_thread::sleep_for( std::chrono::milliseconds( pause ) );
            }

            // bei erfolgreichem Senden keine Wiederholung
            if ( i < repeat ) 
            {
                break;
            }
        }

        // falls das Limit erreicht wurde, konnte der Befehl nicht erfolgreich gesendet werden!
        if ( limit == 0 ) 
        {
            return -IV_ERROR_UNREACHABLE;
        }

        memcpy( buffer, (unsigned char *)&receive[8], (count > bufSize) ? bufSize : count );

        buffer += bufSize;
        address += bufSize;
        count -= ((count > bufSize) ? bufSize : count);
    }

    // Zeitpunkt der letzten Timeboykommunikation halten
    if ( (address & 0x80000000) == 0 )
    {
        m_timeboy_slot[device_address_receiver()].set_receive_stamp_timeboy( clock() );
    }

    return res;
}
